from django.urls import reverse_lazy
from django.utils.functional import cached_property

from dataclasses import dataclass


@dataclass
class AdminUrlLookup:
    url_name: str
    args: tuple


class AdminUrlMixin(object):
    """
    Mixin which can be used to add API to a model for use by standard admin
    component templates, tags, and filters.
    """

    def _get_admin_namespace(self):
        raise NotImplementedError

    def _get_url_args(self):
        raise NotImplementedError

    def _get_url_names(self):
        # Default view names - extend on your model to add extra views
        return ["add", "edit", "delete", "perms"]

    @cached_property
    def urls(self):
        return {
            view: reverse_lazy(lookup.url_name, args=lookup.args)
            for view, lookup in self.url_names.items()
        }

    @cached_property
    def url_names(self) -> dict[str, AdminUrlLookup]:
        namespace = self._get_admin_namespace()
        args = self._get_url_args()
        crud = {
            name: AdminUrlLookup(
                url_name=f"{namespace}:{name}",
                args=[each for each in args if each is not None],
            )
            for name in self._get_url_names()
        }
        return crud



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.utils.translation import gettext_lazy as _
from modelforms.forms import ModelForm

from touchtechnology.common.forms.mixins import UserMixin

UserModel = get_user_model()


class UserEditForm(UserMixin, ModelForm):
    class Meta:
        model = UserModel
        fields = "__all__"
        help_texts = {
            "is_active": _("A disabled account cannot be used to login to the site."),
            "is_staff": _(
                "Content editor's have access to this administration interface."
            ),
            "is_superuser": _("Should this user be all knowing and all powerful?"),
        }
        labels = {
            "is_active": _("Enabled"),
            "is_staff": _("Content editor"),
            "is_superuser": _("God mode"),
        }


class GroupEditForm(UserMixin, ModelForm):
    def __init__(self, *args, **kwargs):
        super(GroupEditForm, self).__init__(*args, **kwargs)
        # The django.contrib applications don't benefit from the
        # touchtechnology.common database fields, so we need to tweak this
        # field queryset the old fashioned way to make sure we're doing a join.
        self.fields["permissions"].queryset = self.fields[
            "permissions"
        ].queryset.select_related()

    class Meta:
        model = Group
        fields = (
            "name",
            "permissions",
        )
        help_texts = {
            "permissions": _("Members of this group will have " "these permissions."),
        }



import logging
import socket
import sys

import django
from django.utils import timezone
from django.utils.deprecation import MiddlewareMixin

from touchtechnology.common.utils import get_timezone_from_request

logger = logging.getLogger(__name__)


class AcceptsMiddleware(MiddlewareMixin):
    def process_request(self, request):
        request.accepts = set(request.META.get("HTTP_ACCEPT", "").split(","))


def served_by_middleware(get_response):
    """
    Expose in our HTTP response headers a number of useful environment values
    that may be useful during debugging.
    """

    def middleware(request):
        response = get_response(request)
        if "HTTP_X_DETAIL" in request.META or response.status_code >= 500:
            response["X-Django-Version"] = django.get_version()
            response["X-Python-Version"] = " ".join(sys.version.split())
            response["X-Served-By"] = socket.gethostname()
            response["X-Timezone-Name"] = timezone.get_current_timezone_name()
        return response

    return middleware


def timezone_middleware(get_response):
    """
    Activate the timezone for this request/response cycle if
    get_timezone_from_request returns a tzinfo.
    """

    def middleware(request):
        tzinfo = get_timezone_from_request(request)

        if tzinfo is not None:
            logger.debug('Activating user timezone "%s" for this request.', tzinfo)
            request.tzinfo = tzinfo
            timezone.activate(tzinfo)

        response = get_response(request)

        if tzinfo is not None:
            logger.debug(
                'Forcibly reset the timezone from "%s" back to "%s".',
                request.tzinfo,
                timezone.get_default_timezone_name(),
            )
            timezone.deactivate()

        return response

    return middleware



import logging
from importlib import metadata

from django import forms
from django.forms.widgets import DateInput, MultiWidget, Select, TimeInput
from django.utils import timezone
from froala_editor.widgets import FroalaEditor
from timezone_field.backends import get_tz_backend
from timezone_field.choices import with_gmt_offset

logger = logging.getLogger(__name__)


class HTMLWidget(FroalaEditor):
    def trigger_froala(self, *args, **kwargs):
        """
        We don't want to do inline scripts, we'll handle the loading once from
        froala.js, so return an empty string to overload the base FroalaEditor
        implementation.

        Note: required for django-froala-editor<3 AFAICT
        """
        version = metadata.version("django-froala-editor")
        if int(version.split(".")[0]) < 3:
            return ""
        return super().trigger_froala(*args, **kwargs)

    class Media:
        css = {
            "all": (
                "https://cdnjs.cloudflare.com/ajax/libs/codemirror/5.3.0/"
                "codemirror.min.css",
                "touchtechnology/common/css/froala_themes/dark.min.css",
                "touchtechnology/common/css/froala_themes/gray.min.css",
                "touchtechnology/common/css/froala_themes/red.min.css",
                "touchtechnology/common/css/froala_themes/royal.min.css",
            ),
        }
        js = (
            "https://cdnjs.cloudflare.com/ajax/libs/codemirror/5.3.0/codemirror.min.js",
            "touchtechnology/common/js/froala.js",
        )


class GoogleMapsWidget(MultiWidget):
    def __init__(self, height, width, zoom, attrs=None):
        if attrs is None:
            attrs = {}
        self.attrs = attrs
        self.height = height
        self.width = width
        self.zoom = zoom
        widgets = (
            forms.TextInput(attrs=dict(placeholder="Latitude", **attrs)),
            forms.TextInput(attrs=dict(placeholder="Longitude", **attrs)),
            forms.TextInput(attrs=dict(placeholder="Zoom", **attrs)),
        )
        super().__init__(widgets, attrs)

    def decompress(self, value):
        if value:
            return value.split(",")
        return ("", "", "")


class SelectDateTimeWidget(MultiWidget):
    def __init__(
        self,
        attrs=None,
        date_format=None,
        time_format=None,
        date_attrs=None,
        time_attrs=None,
        use_pytz=False,
    ):
        tz_backend = get_tz_backend(use_pytz=use_pytz)
        tz_choices = [("", "---")] + with_gmt_offset(
            tz_backend.base_tzstrs, use_pytz=use_pytz
        )
        widgets = (
            DateInput(
                attrs=attrs if date_attrs is None else date_attrs,
                format=date_format,
            ),
            TimeInput(
                attrs=attrs if time_attrs is None else time_attrs,
                format=time_format,
            ),
            Select(attrs=attrs, choices=tz_choices),
        )
        super().__init__(widgets)

    def decompress(self, value):
        if value:
            value = value.astimezone(timezone.get_current_timezone())
            return [value.date(), value.time(), value.tzinfo]
        return [None, None, None]


class SelectDateTimeHiddenWidget(SelectDateTimeWidget):
    def __init__(self, attrs=None):
        widgets = (
            forms.HiddenInput(),
            forms.HiddenInput(),
            forms.HiddenInput(),
        )
        super().__init__(widgets, attrs)



import os.path
import tempfile

from django.test import TestCase

from touchtechnology.common.storage import FileSystemStorage


class StorageTests(TestCase):

    def setUp(self):
        location = tempfile.mkdtemp()
        self.storage = FileSystemStorage(location=location)

    def test_makedirs(self):
        path, folder = "", "test_mkdir_folder/with_two/subfolders"
        self.assertEqual(self.storage.listdir(path), ([], []))
        self.storage.makedirs(folder)
        self.assertEqual(
            self.storage.listdir(os.path.dirname(folder)),
            ([os.path.basename(folder)], []),
        )
        self.assertEqual(
            self.storage.listdir(os.path.dirname(os.path.dirname(folder))),
            ([os.path.basename(os.path.dirname(folder))], []),
        )

    def test_mkdir(self):
        path, folder = "", "test_mkdir_folder"
        self.assertEqual(self.storage.listdir(path), ([], []))
        self.storage.mkdir(folder)
        self.assertEqual(self.storage.listdir(path), ([folder], []))
        self.assertEqual(self.storage.listdir(folder), ([], []))

    def test_mkdir_path(self):
        path, folder = "path", "test_mkdir_path_folder"
        self.storage.mkdir(path)
        self.assertEqual(self.storage.listdir(path), ([], []))
        self.storage.mkdir(folder, path)
        self.assertEqual(self.storage.listdir(path), ([folder], []))
        self.assertEqual(self.storage.listdir(os.path.join(path, folder)), ([], []))

    def test_path(self):
        filename = "test_path/to/file.txt"
        self.assertEqual(
            self.storage.path(filename), os.path.join(self.storage.location, filename)
        )

    def test_rmdir(self):
        path, folder = "", "test_rmdir"
        with self.assertRaises(OSError) as cm:
            self.storage.rmdir(folder)
        self.assertEqual(cm.exception.errno, 2)
        self.storage.mkdir(folder)
        self.assertEqual(self.storage.listdir(path), ([folder], []))
        self.storage.rmdir(folder)
        self.assertEqual(self.storage.listdir(path), ([], []))

    def test_rmdir_subdir(self):
        folder = "test_rmdir_subdir"
        subdir = os.path.join(folder, "subdir")
        self.storage.makedirs(subdir)
        with self.assertRaises(OSError):
            self.storage.rmdir(folder)

    def test_writable(self):
        folder = "test_writable"
        with self.assertRaises(OSError) as cm:
            self.storage.writable(folder)
        self.assertEqual(cm.exception.errno, 2)
        self.storage.mkdir(folder)
        self.assertTrue(self.storage.writable(folder))



from datetime import timedelta
from itertools import zip_longest

from dateutil.parser import parse
from dateutil.rrule import DAILY, WEEKLY
from django import template
from django.apps import apps
from django.db.models import Case, F, Q, Sum, When
from django.template.loader import get_template
from django.utils import timezone
from first import first

register = template.Library()


@register.filter
def get_competitions(node):
    return node.get_children().filter(content_type__model="competition")


@register.filter
def opponent(match, team):
    if not match:
        return None
    elif match.home_team == team:
        return match.get_away_team()
    elif match.away_team == team:
        return match.get_home_team()
    else:
        return None


@register.inclusion_tag(
    "tournamentcontrol/competition/next_date.html", takes_context=True
)
def next_date(context, season, offset=0, datestr=None):
    from tournamentcontrol.competition.models import Match, Season

    if isinstance(season, str):
        season = Season.objects.get(pk=season)

    # TODO make the offset a value stored on the competition
    if datestr is None:
        now = timezone.now() - timedelta(minutes=offset)
    else:
        now = parse(datestr) - timedelta(minutes=offset)

    # drop any signifigance less than the minute
    now = now.replace(second=0, microsecond=0)

    # start with just matches for this season (and by definition competition)
    matches = Match.objects.filter(
        stage__division__draft=False, stage__division__season=season
    ).select_related(
        "play_at",
        "home_team__club",
        "away_team__club",
        "home_team__division__season__competition",
        "away_team__division__season__competition",
        "stage__division__season__competition",
    )

    # restrict to matches starting after "now"
    later_today = Q(date__exact=now.date(), time__gte=now.time())
    tomorrow_onwards = Q(date__gt=now.date())
    matches = matches.filter(later_today | tomorrow_onwards)

    try:
        next_game_date = first(matches.dates("date", "day").order_by("date"))
    except Match.DoesNotExist:
        matches = matches.none()
        next_game_date = None
    else:
        matches = matches.filter(date=next_game_date).exclude(is_bye=True)

    next_round = None

    if season.mode == DAILY:
        # For tournaments we want to update the upcoming matches every timeslot
        try:
            next_round = matches.filter(
                home_team_score__isnull=True, away_team_score__isnull=True
            )[0]
        except IndexError:
            matches = matches.none()
        else:
            matches = matches.filter(date=next_round.date, time=next_round.time)

    elif season.mode == WEEKLY:
        # For a weekly competition we want to update the upcoming matches each
        # week and sorted by division.
        matches = matches.order_by("stage__division", *Match._meta.ordering)

    context.update(
        {
            "matches": matches,
            "next_game_date": next_game_date,
            "next_round": next_round,
            "competition": season.competition,
            "season": season,
        }
    )

    return context


@register.simple_tag(takes_context=True)
def ladder(context, stage):
    if stage.pools.count():
        ordering = stage.ladder_summary.model._meta.ordering
        summary = stage.ladder_summary.order_by("stage_group", *ordering)
        template_name = "tournamentcontrol/competition/ladder/pool.html"
    else:
        summary = stage.ladder_summary.all()
        template_name = "tournamentcontrol/competition/ladder/standard.html"

    context.update(
        {
            "summary": summary,
        }
    )

    tpl = get_template(template_name)
    return tpl.render(context)


@register.simple_tag
def score(match, team, template_name="tournamentcontrol/competition/_score.html"):
    """
    Output a score relative to the specified team. Also includes the 'status'
    of the result from the perspective of the team.

        ie. Win 6-4
            Lose 3-5
            Forfeit 0-6

    Use the third argument to specify an alternative template file name.
    """
    if team == match.home_team:
        team_score = match.home_team_score
        opponent_score = match.away_team_score
    else:
        team_score = match.away_team_score
        opponent_score = match.home_team_score

    if team_score is not None and opponent_score is not None:
        if team_score > opponent_score or team == match.forfeit_winner:
            result = "won"
        elif team_score < opponent_score or match.is_forfeit:
            result = "lost"
        elif team_score == opponent_score:
            result = "drew"
        else:
            raise ValueError(
                "An error has occured while determining the " "match result."
            )
    else:
        result = None

    context = {
        "result": result,
        "team_score": team_score,
        "opponent_score": opponent_score,
        "forfeit": match.is_forfeit,
    }

    t = template.loader.get_template(template_name)

    return t.render(context)


@register.filter
def teams_in_season(team_queryset, season):
    assert isinstance(season, apps.get_model("competition", "season"))
    return team_queryset.filter(division__season=season)


@register.filter
def teams_in_division(team_queryset, division):
    assert isinstance(division, apps.get_model("competition", "division"))
    return team_queryset.filter(division=division)


@register.filter
def players(team, pad_to=None):
    try:
        players = list(team.people.filter(is_player=True))
    except AttributeError:
        players = []
    if pad_to is not None:
        while len(players) < pad_to:
            players.append(None)
    return players


@register.filter
def pair(i1, i2):
    return list(zip_longest(i1, i2))


@register.inclusion_tag("tournamentcontrol/competition/templatetags/statistics.html")
def statistics(match):
    stats = match.statistics.filter(played__gt=0).select_related("player")

    home_stats = stats.filter(
        match__home_team__people__team=match.home_team,
        player__teamassociation__team=match.home_team,
    ).distinct()

    away_stats = stats.filter(
        match__away_team__people__team=match.away_team,
        player__teamassociation__team=match.away_team,
    ).distinct()

    context = {
        "match": match,
        "lines": list(zip_longest(home_stats, away_stats)),
        "home_stats": home_stats,
        "away_stats": away_stats,
    }

    return context


@register.inclusion_tag("tournamentcontrol/competition/templatetags/preview.html")
def preview(match):
    annotate = {
        "played": Sum(
            Case(
                When(
                    person__statistics__match__stage__division=match.stage.division,
                    then=F("person__statistics__played"),
                ),
                default=0,
            )
        ),
        "points": Sum(
            Case(
                When(
                    person__statistics__match__stage__division=match.stage.division,
                    then=F("person__statistics__points"),
                ),
                default=0,
            )
        ),
    }

    home_team = match.home_team.people.filter(is_player=True).annotate(**annotate)
    away_team = match.away_team.people.filter(is_player=True).annotate(**annotate)

    context = {
        "match": match,
        "lines": list(zip_longest(home_team, away_team)),
        "home_team": home_team,
        "away_team": away_team,
    }

    return context



import json
import re
from datetime import datetime

from django import forms
from django.conf import settings
from django.db.models import Min
from django.utils import timezone
from django.utils.encoding import smart_str
from django.utils.translation import gettext_lazy as _
from mptt.forms import TreeNodeChoiceField
from namedentities import named_entities
from timezone_field import TimeZoneFormField

from touchtechnology.common.forms.iter import TemplateChoiceIterator
from touchtechnology.common.forms.mixins import LabelFromInstanceMixin
from touchtechnology.common.forms.widgets import (
    GoogleMapsWidget,
    HTMLWidget,
    SelectDateTimeHiddenWidget,
    SelectDateTimeWidget,
)


class EmailField(forms.EmailField):
    """
    Custom EmailField which will transform email address to lowercase unless
    advised otherwise as this SHOULD NOT be significant.
    """

    def __init__(self, lowercase=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lowercase = lowercase

    def clean(self, data):
        data = super().clean(data)
        if self.lowercase:
            data = data.lower()
        return data


class ModelMultipleChoiceField(LabelFromInstanceMixin, forms.ModelMultipleChoiceField):
    def __init__(self, select_related=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if select_related and self.queryset is not None:
            self.queryset = self.queryset.select_related()


class HTMLField(forms.CharField):
    widget = HTMLWidget

    def widget_attrs(self, widget):
        options = getattr(settings, "FROALA_EDITOR_OPTIONS", {})
        return {"data-options": json.dumps(options)}

    def clean(self, value):
        if isinstance(value, str):
            value = named_entities(value)
        return super().clean(value)


class GoogleMapsField(forms.MultiValueField):
    def __init__(self, max_length, width=300, height=200, zoom=8, *args, **kwargs):
        fields = (
            forms.CharField(max_length=25, required=False),
            forms.CharField(max_length=25, required=False),
            forms.IntegerField(required=False),
        )
        kwargs["widget"] = GoogleMapsWidget(height, width, zoom)
        super().__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        return ",".join([smart_str(each) for each in data_list])

    def clean(self, data):
        if [d for d in data if d]:
            return ",".join(data)
        return super().clean(data)


class SelectDateTimeField(forms.MultiValueField):
    widget = SelectDateTimeWidget
    hidden_widget = SelectDateTimeHiddenWidget
    default_error_messages = {
        "invalid_date": _("Enter a valid date."),
        "invalid_time": _("Enter a valid time."),
        "invalid_tz": _("Enter a valid time zone."),
    }

    def __init__(self, *, input_date_formats=None, input_time_formats=None, **kwargs):
        errors = self.default_error_messages.copy()
        if "error_messages" in kwargs:
            errors.update(kwargs["error_messages"])
        localize = kwargs.get("localize", False)
        fields = (
            forms.DateField(
                input_formats=input_date_formats,
                error_messages={"invalid": errors["invalid_date"]},
                localize=localize,
            ),
            forms.TimeField(
                input_formats=input_time_formats,
                error_messages={"invalid": errors["invalid_time"]},
                localize=localize,
            ),
            TimeZoneFormField(
                use_pytz=False,
                choices_display="WITH_GMT_OFFSET",
                error_messages={"required": errors["invalid_tz"]},
            ),
        )
        super().__init__(fields, **kwargs)

    def compress(self, data_list):
        if data_list:
            # Raise a validation error if time or date is empty
            # (possible if SplitDateTimeField has required=False).
            if data_list[0] in self.empty_values:
                raise forms.ValidationError(
                    self.error_messages["invalid_date"], code="invalid_date"
                )
            if data_list[1] in self.empty_values:
                raise forms.ValidationError(
                    self.error_messages["invalid_time"], code="invalid_time"
                )
            if data_list[2] in self.empty_values:
                raise forms.ValidationError(
                    self.error_messages["invalid_tz"], code="invalid_tz"
                )
            result = datetime.combine(*data_list[:-1])
            return timezone.make_aware(result, data_list[-1])
        return None


class TemplatePathFormField(forms.ChoiceField):
    def __init__(
        self,
        template_base,
        template_folder,
        match=None,
        recursive=False,
        required=True,
        widget=None,
        label=None,
        initial=None,
        help_text=None,
        cache_choices=False,
        empty_label=_("Default"),
        # This will never be used, but it's now a CharField in the
        # database so we need to pull it out of the keyword
        # arguments.
        max_length=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            required=required,
            widget=widget,
            label=label,
            initial=initial,
            help_text=help_text,
            *args,
            **kwargs,
        )

        self.template_base = template_base
        self.template_folder = template_folder
        self.match = match
        self.recursive = recursive
        self.empty_label = empty_label

        if self.match is not None:
            self.match_re = re.compile(self.match)

        if not self.template_base.endswith("/"):
            self.template_base += "/"

        self.widget.choices = self.choices

    def _get_choices(self):
        return TemplateChoiceIterator(self)

    choices = property(_get_choices, forms.ChoiceField.choices.fset)


class ModelChoiceField(LabelFromInstanceMixin, forms.ModelChoiceField):
    def __init__(self, select_related=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if select_related and self.queryset is not None:
            self.queryset = self.queryset.select_related()


class SitemapNodeModelChoiceField(ModelChoiceField):
    def __init__(self, label_from_instance="title", *args, **kwargs):
        super().__init__(label_from_instance=label_from_instance, *args, **kwargs)


def boolean_choice_field_coerce(value):
    return bool(int(value))


class MinTreeNodeChoiceField(TreeNodeChoiceField):
    @property
    def minimum_level(self):
        if not hasattr(self, "_minimum_level"):
            self._minimum_level = self.queryset.aggregate(minimum=Min("level")).get(
                "minimum"
            )
        return self._minimum_level

    def label_from_instance(self, obj):
        return f"{self.level_indicator * (obj.level - self.minimum_level)} {smart_str(obj)}"



# -*- coding: utf-8 -*-

import django.db.models.deletion
from django.db import migrations, models

import tournamentcontrol.competition.models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0016_season_forfeit_notifications"),
    ]

    operations = [
        migrations.AddField(
            model_name="competition",
            name="rank_importance",
            field=models.DecimalField(
                max_digits=6, decimal_places=3, blank=True, null=True
            ),
        ),
        migrations.AddField(
            model_name="season",
            name="rank_importance",
            field=models.DecimalField(
                max_digits=6, decimal_places=3, blank=True, null=True
            ),
        ),
        migrations.AddField(
            model_name="division",
            name="rank_importance",
            field=models.DecimalField(
                max_digits=6, decimal_places=3, blank=True, null=True
            ),
        ),
        migrations.AddField(
            model_name="stage",
            name="rank_importance",
            field=models.DecimalField(
                max_digits=6, decimal_places=3, blank=True, null=True
            ),
        ),
        migrations.AddField(
            model_name="stagegroup",
            name="rank_importance",
            field=models.DecimalField(
                max_digits=6, decimal_places=3, blank=True, null=True
            ),
        ),
        migrations.AddField(
            model_name="match",
            name="rank_importance",
            field=models.DecimalField(
                max_digits=6, decimal_places=3, blank=True, null=True
            ),
        ),
        migrations.CreateModel(
            name="RankDivision",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("title", models.CharField(verbose_name="Title", max_length=255)),
                (
                    "short_title",
                    models.CharField(
                        verbose_name="Short title",
                        blank=True,
                        help_text="This is used in navigation menus instead of the longer title value.",
                        max_length=100,
                    ),
                ),
                ("slug", models.SlugField(verbose_name="Slug", max_length=255)),
                (
                    "slug_locked",
                    models.BooleanField(verbose_name="Slug locked", default=False),
                ),
                ("order", models.PositiveIntegerField(default=1)),
                ("enabled", models.BooleanField(default=True)),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
            bases=(tournamentcontrol.competition.models.AdminUrlMixin, models.Model),
        ),
        migrations.CreateModel(
            name="RankPoints",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "points",
                    models.DecimalField(default=0, max_digits=6, decimal_places=3),
                ),
                ("date", models.DateField()),
            ],
            options={
                "ordering": ("-date", "team", "-points"),
            },
        ),
        migrations.CreateModel(
            name="RankTeam",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "ordering": ("division",),
            },
        ),
        migrations.AddField(
            model_name="division",
            name="rank_division",
            field=models.ForeignKey(
                null=True,
                blank=True,
                to="competition.RankDivision",
                on_delete=django.db.models.deletion.PROTECT,
            ),
        ),
        migrations.AddField(
            model_name="team",
            name="rank_division",
            field=models.ForeignKey(
                null=True,
                blank=True,
                to="competition.RankDivision",
                on_delete=django.db.models.deletion.PROTECT,
            ),
        ),
        migrations.AddField(
            model_name="rankteam",
            name="club",
            field=models.ForeignKey(
                to="competition.Club", on_delete=django.db.models.deletion.PROTECT
            ),
        ),
        migrations.AddField(
            model_name="rankteam",
            name="division",
            field=models.ForeignKey(
                to="competition.RankDivision",
                on_delete=django.db.models.deletion.PROTECT,
            ),
        ),
        migrations.AlterUniqueTogether(
            name="rankteam",
            unique_together=set([("club", "division")]),
        ),
        migrations.AddField(
            model_name="rankpoints",
            name="team",
            field=models.ForeignKey(
                to="competition.RankTeam", on_delete=django.db.models.deletion.PROTECT
            ),
        ),
    ]



from rest_framework import serializers
from rest_framework_nested.serializers import NestedHyperlinkedModelSerializer

from tournamentcontrol.competition import models

from .viewsets import SlugViewSet


class ListSeasonSerializer(NestedHyperlinkedModelSerializer):
    parent_lookup_kwargs = {"competition_slug": "competition__slug"}

    class Meta:
        model = models.Season
        fields = ("title", "slug", "url")
        extra_kwargs = {"url": {"lookup_field": "slug"}}


class ListCompetitionSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.Competition
        fields = ("title", "slug", "url")
        extra_kwargs = {"url": {"lookup_field": "slug"}}


class CompetitionSerializer(ListCompetitionSerializer):
    seasons = ListSeasonSerializer(many=True, read_only=True)

    class Meta(ListCompetitionSerializer.Meta):
        fields = ("title", "slug", "url", "seasons")


class CompetitionViewSet(SlugViewSet):
    queryset = models.Competition.objects.all()
    serializer_class = CompetitionSerializer
    list_serializer_class = ListCompetitionSerializer



from django.conf import settings
from first import first

__all__ = (
    "NODE_CACHE",
    "PAGE_CONTENT_BLOCKS",
    "PAGE_CONTENT_CLASSES",
    "PAGE_TEMPLATE_BASE",
    "PAGE_TEMPLATE_FOLDER",
    "PAGE_TEMPLATE_REGEX",
)


def S(name, default=None):
    return getattr(settings, "TOUCHTECHNOLOGY_" + name, default)


project_template_dirs = first(getattr(settings, "TEMPLATES", ()), {}).get("DIRS", [])
project_template_base = first(project_template_dirs, "templates")

NODE_CACHE = getattr(settings, "TOUCHTECHNOLOGY_NODE_CACHE", "default")

# These settings allow us to determine where page templates live, which allows
# us not to need to specify each actual template in code.

PAGE_CONTENT_BLOCKS = S("PAGE_CONTENT_BLOCKS", 1)
PAGE_CONTENT_CLASSES = S("PAGE_CONTENT_CLASSES", ("copy",))
PAGE_TEMPLATE_BASE = S("PAGE_TEMPLATE_BASE", project_template_base)
PAGE_TEMPLATE_FOLDER = S("PAGE_TEMPLATE_FOLDER", "touchtechnology/content/pages/")
PAGE_TEMPLATE_REGEX = S("PAGE_TEMPLATE_REGEX", r"\.html$")

TENANT_MEDIA_PUBLIC = S("TENANT_MEDIA_PUBLIC", True)



# -*- coding: utf-8 -*-

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0027_matchvideo_data"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="matchvideo",
            name="match",
        ),
        migrations.DeleteModel(
            name="MatchVideo",
        ),
    ]



import logging
import os

from django.contrib import messages
from django.core.files.storage import default_storage
from django.db.models import Q
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import include, path, re_path, resolve, reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from first import first

from touchtechnology.admin.base import AdminComponent
from touchtechnology.admin.mixins import AdminUrlMixin
from touchtechnology.common.decorators import (
    require_POST_m,
    staff_login_required_m,
)
from touchtechnology.common.models import SitemapNode
from touchtechnology.content.forms import (
    FileUploadForm,
    NewFolderForm,
    NewPlaceholderSitemapNodeForm,
    NewSitemapNodeForm,
    Page,
    PageContentFormset,
    PageForm,
    PlaceholderConfigurationBase,
    PlaceholderContentFormset,
    PlaceholderSitemapNodeForm,
    RedirectEditForm,
    SitemapNodeForm,
)
from touchtechnology.content.models import (
    Chunk,
    PageContent,
    Placeholder,
    Redirect,
)

logger = logging.getLogger(__name__)


class ContentAdminComponent(AdminComponent):
    verbose_name = _("Content")

    filter = Q(content_type__app_label="content") | Q(content_type__isnull=True)

    def __init__(self, app, name="content", app_name="content"):
        super(ContentAdminComponent, self).__init__(app, name, app_name)

    def get_urls(self):
        folder_patterns = (
            [
                path("add/", self.edit_folder, name="add"),
                path("<int:pk>/", self.edit_folder, name="edit"),
                path("<int:pk>/delete/", self.delete_folder, name="delete"),
            ],
            self.app_name,
        )

        page_patterns = (
            [
                path("add/", self.edit_page, name="add"),
                path("<int:pk>/", self.edit_page, name="edit"),
                path("<int:pk>/delete/", self.delete_page, name="delete"),
                path("<int:pk>/permission/", self.perms_sitemapnode, name="perms"),
            ],
            self.app_name,
        )

        redirect_patterns = (
            [
                path("", self.list_redirects, name="list"),
                path("create/", self.edit_redirect, name="add"),
                path("<int:pk>/", self.edit_redirect, name="edit"),
                path("<int:pk>/delete/", self.delete_redirect, name="delete"),
                path("<int:pk>/permission/", self.perms_redirect, name="perms"),
            ],
            self.app_name,
        )

        placeholder_patterns = (
            [
                path("", self.list_placeholders, name="list"),
                path("create/", self.edit_placeholder_x, name="add"),
                path("<int:pk>/", self.edit_placeholder_x, name="edit"),
                path("<int:pk>/delete/", self.delete_placeholder_x, name="delete"),
                # path("<int:pk>/permission/", self.perms_placeholder, name="delete"),
            ],
            self.app_name,
        )

        application_patterns = (
            [
                path("add/", self.edit_application_node, name="add"),
                path("<int:pk>/", self.edit_application_node, name="edit"),
                path("<int:pk>/delete/", self.delete_application_node, name="delete"),
                path("<int:pk>/permission/", self.perms_sitemapnode, name="perms"),
            ],
            self.app_name,
        )

        chunk_patterns = (
            [
                path("", self.list_chunks, name="list"),
                path("create/", self.edit_chunk, name="add"),
                path("<int:pk>/", self.edit_chunk, name="edit"),
                path("<int:pk>/delete/", self.delete_chunk, name="delete"),
                path("<int:pk>/permission/", self.perms_chunk, name="perms"),
            ],
            self.app_name,
        )

        files_patterns = (
            [
                re_path(
                    r"^(?P<path>.+/)?(?P<filename>[^/]+?)/rm/$", self.rm, name="delete"
                ),
                re_path(r"^(?P<path>.+/)rmdir/$", self.rm, name="delete"),
                path("", self.ls, name="list"),
                re_path(r"^(?P<path>.+)$", self.ls, name="edit"),
            ],
            self.app_name,
        )

        urlpatterns = [
            path("", self.list_pages, name="index"),
            path("folder/", include(folder_patterns, namespace="folder")),
            path("page/", include(page_patterns, namespace="page")),
            path("redirect/", include(redirect_patterns, namespace="redirect")),
            path(
                "placeholder/",
                include(placeholder_patterns, namespace="real_placeholder"),
            ),
            path(
                "application/", include(application_patterns, namespace="placeholder")
            ),
            path("chunk/", include(chunk_patterns, namespace="chunk")),
            path("files/", include(files_patterns, namespace="media")),
        ]
        return urlpatterns

    def dropdowns(self):
        dl = (
            (_("Sitemap"), self.reverse("index"), "sitemap"),
            (_("Media"), self.reverse("media:list"), "folder-open"),
            (_("Chunks"), self.reverse("chunk:list"), "pencil-square-o"),
            (_("Redirects"), self.reverse("redirect:list"), "reply"),
            (_("Applications"), self.reverse("real_placeholder:list"), "puzzle-piece"),
        )
        return dl

    # Generic views - trial before moving onto Application

    def generic_parent_edit_related(
        self,
        request,
        pk,
        parent_model,
        child_model,
        child_related_model,
        parent_form_class=None,
        child_form_class=None,
        child_formset_class=None,
        post_save_redirect=None,
        *args,
        **kwargs,
    ):
        if pk is not None:
            instance = get_object_or_404(parent_model, pk=pk)
        else:
            instance = None
        if post_save_redirect is None:
            post_save_redirect = self.redirect(self.reverse("index"))
        return self.generic_edit_related(
            request,
            child_model,
            child_related_model,
            instance=instance,
            form_class=child_form_class,
            form_kwargs={
                "parent_form_class": parent_form_class,
                "parent_form_kwargs": {"user": request.user},
            },
            formset_class=child_formset_class,
            post_save_redirect=post_save_redirect,
            extra_context=kwargs,
        )

    # Shared perms view

    def perms_sitemapnode(self, request, pk, **extra_context):
        return self.generic_permissions(request, SitemapNode, pk=pk, **extra_context)

    # Folder views -- folders are just SitemapNode's with no links

    @staff_login_required_m
    def edit_folder(self, request, pk=None, *args, **kwargs):
        form_class = NewSitemapNodeForm if pk is None else SitemapNodeForm
        return self.generic_edit(
            request,
            SitemapNode,
            pk=pk,
            form_class=form_class,
            related=(),
            permission_required=True,
            extra_context=kwargs,
        )

    @staff_login_required_m
    def delete_folder(self, request, pk, *args, **kwargs):
        node = get_object_or_404(SitemapNode, pk=pk)
        if not node.is_leaf_node():
            raise Http404
        return self.generic_delete(request, SitemapNode, pk=pk, extra_context=kwargs)

    # Page views

    @staff_login_required_m
    def list_pages(self, request, **extra_context):
        try:
            root = SitemapNode.objects.get(pk=request.GET["r"])
            descendants = root.get_descendants()
            nodes = descendants.filter(self.filter).filter(level=root.level + 1)
        except KeyError:
            nodes = SitemapNode.objects.filter(self.filter).filter(level=0)
        return self.generic_list(
            request,
            nodes,
            paginate_by=0,
            permission_required=True,
            extra_context=extra_context,
        )

    @staff_login_required_m
    def edit_page(self, request, pk=None, **extra_context):
        post_save_redirect = None
        if pk is not None:
            instance = get_object_or_404(SitemapNode, pk=pk)
            if instance.parent_id:
                post_save_redirect = self.redirect(
                    self.reverse("index") + "?r=%d" % instance.parent_id
                )
        return self.generic_parent_edit_related(
            request,
            pk,
            parent_model=SitemapNode,
            child_model=Page,
            child_related_model=PageContent,
            parent_form_class=SitemapNodeForm,
            child_form_class=PageForm,
            child_formset_class=PageContentFormset,
            post_save_redirect=post_save_redirect,
            **extra_context,
        )

    @staff_login_required_m
    def delete_page(self, request, pk, *args, **kwargs):
        node = get_object_or_404(SitemapNode, pk=pk)

        if request.method == "POST":
            # verify that we are only operating on a leaf node
            if not node.is_leaf_node():
                raise Http404
            # if this was not a POST request, we don't make changes to the
            # database, we just display the warning page.
            page = node.object
            page.content.all().delete()
            page.delete()
            node.delete()
            redirect_to = self.reverse("index")
            if node.parent:
                redirect_to += "?r=%d" % node.parent_id
            return HttpResponseRedirect(redirect_to)

        context = {
            "object": node,
        }
        context.update(kwargs)

        templates = self.template_path("delete_confirm.html")
        return self.render(request, templates, context)

    # Redirect views

    @staff_login_required_m
    def list_redirects(self, request, *args, **kwargs):
        return self.generic_list(
            request, Redirect, permission_required=True, extra_context=kwargs
        )

    @staff_login_required_m
    def edit_redirect(self, request, pk=None, *args, **kwargs):
        return self.generic_edit(
            request,
            Redirect,
            pk=pk,
            form_class=RedirectEditForm,
            permission_required=True,
            extra_context=kwargs,
        )

    @staff_login_required_m
    def delete_redirect(self, request, pk, *args, **kwargs):
        return self.generic_delete(
            request, Redirect, pk=pk, permission_required=True, extra_context=kwargs
        )

    @staff_login_required_m
    def perms_redirect(self, request, pk, *args, **kwargs):
        return self.generic_permissions(request, Redirect, pk=pk, **kwargs)

    # Application views

    @staff_login_required_m
    def list_placeholders(self, request, *args, **kwargs):
        return self.generic_list(
            request,
            Placeholder,
            paginate_by=0,
            permission_required=True,
            extra_context=kwargs,
        )

    @staff_login_required_m
    def edit_placeholder_x(self, request, pk=None, *args, **kwargs):
        return self.generic_edit(
            request, Placeholder, pk=pk, permission_required=True, extra_context=kwargs
        )

    @staff_login_required_m
    def delete_placeholder_x(self, request, pk, *args, **kwargs):
        return self.generic_delete(
            request, Placeholder, pk=pk, permission_required=True, extra_context=kwargs
        )

    @staff_login_required_m
    def edit_application_node(self, request, pk=None, **extra_context):
        kwargs_form_class = PlaceholderConfigurationBase

        if pk is None:
            instance = SitemapNode()
            form_class = NewPlaceholderSitemapNodeForm
        else:
            instance = get_object_or_404(SitemapNode, pk=pk)
            form_class = PlaceholderSitemapNodeForm
            try:
                site = instance.object.site(instance)
            except ValueError:
                logger.exception("Application is unavailable, " "no custom form class.")
            else:
                kwargs_form_class = site.kwargs_form_class or kwargs_form_class

        if request.method == "POST":
            form = form_class(data=request.POST, instance=instance, user=request.user)
            kw = kwargs_form_class(
                data=request.POST, instance=instance, prefix="kw", user=request.user
            )
            copy = PlaceholderContentFormset(
                data=request.POST, instance=instance, prefix="copy"
            )
            if form.is_valid() and kw.is_valid() and copy.is_valid():
                form.save()
                kw.save()
                copy.save()

                redirect_to = self.reverse("index")
                if instance.parent:
                    redirect_to += "?r=%d" % instance.parent_id

                # post_save_callback if we ever move this to a generic_edit
                # pattern in future.
                instance.object.invalidate_cache()

                return HttpResponseRedirect(redirect_to)
        else:
            form = form_class(instance=instance, user=request.user)
            kw = kwargs_form_class(instance=instance, prefix="kw", user=request.user)
            copy = PlaceholderContentFormset(instance=instance, prefix="copy")

        context = {
            "form": form,
            "formsets": {
                "kwargs": kw,
                "copy": copy,
            },
            "formset_media": kw.media + copy.media,
            "node": instance,
            "model": Placeholder.objects.none(),
        }
        context.update(extra_context)

        templates = self.template_path("edit_application.html")
        return self.render(request, templates, context)

    @require_POST_m
    @staff_login_required_m
    def delete_application_node(self, request, pk, *args, **kwargs):
        node = get_object_or_404(SitemapNode, pk=pk)

        if request.method == "POST":
            # verify that we are only operating on a leaf node
            if not node.is_leaf_node():
                raise Http404
            redirect_to = reverse("admin:content:index", current_app=self.app)
            if node.parent:
                redirect_to += "?r=%d" % node.parent_id
            # if this was not a POST request, we don't make changes to the
            # database, we just display the warning page.
            node.delete()

            return HttpResponseRedirect(redirect_to)

        context = {
            "object": node,
        }
        context.update(kwargs)

        templates = self.template_path("delete_confirm.html")
        return self.render(request, templates, context)

    # Chunk views

    @staff_login_required_m
    def list_chunks(self, request, *args, **kwargs):
        return self.generic_list(request, Chunk, extra_context=kwargs)

    @staff_login_required_m
    def edit_chunk(self, request, pk=None, *args, **kwargs):
        return self.generic_edit(request, Chunk, pk=pk, extra_context=kwargs)

    @staff_login_required_m
    def delete_chunk(self, request, pk, **extra_context):
        return self.generic_delete(request, Chunk, pk=pk, extra_context=extra_context)

    @staff_login_required_m
    def perms_chunk(self, request, pk, *args, **kwargs):
        return self.generic_permissions(request, Chunk, pk=pk, **kwargs)

    @staff_login_required_m
    def ls(self, request, path=None, *args, **kwargs):
        try:
            directories, files = default_storage.listdir(path or ".")
        except OSError as e:
            raise Http404(e)

        directories.sort()
        files.sort()

        parentdir = os.path.dirname(path[:-1]) or "ROOT" if path else ""

        writable = True  # presume it is, and fail gracefully if it isn't
        if path and hasattr(default_storage, "writable"):
            writable = default_storage.writable(path)

        # Both types for Form expect a path to be passed in, set up the base.
        form_kwargs = dict(path=path)

        if request.method == "POST":
            # Give the Forms the POST and FILES from the request
            form_kwargs.update(dict(data=request.POST, files=request.FILES))

            folder_form = NewFolderForm(prefix="folder", **form_kwargs)
            file_form = FileUploadForm(prefix="file", **form_kwargs)

            if folder_form.is_valid() and file_form.is_valid():
                redirect_to = "."
                saved = None

                if folder_form.has_changed():
                    saved = folder_form.save()
                    logger.debug("created folder %s", saved)
                    folder_name = os.path.basename(saved)
                    messages.success(
                        request, _('Opened new folder "%s".') % folder_name
                    )
                    redirect_to = folder_name + "/"  # FIXME reverse this

                if file_form.has_changed():
                    saved = file_form.save()
                    file_name = os.path.basename(saved)
                    messages.success(request, _('Uploaded new file "%s"') % file_name)

                if saved is None:
                    messages.warning(request, _("There were no changes to be saved."))

                return self.redirect(redirect_to)

        else:
            folder_form = NewFolderForm(prefix="folder", **form_kwargs)
            file_form = FileUploadForm(prefix="file", **form_kwargs)

        def _directories(d):
            fullpath = os.path.join(path or "", d, "")
            return fullpath, d, default_storage.listdir(fullpath)

        directories = [_directories(d) for d in directories]
        match = resolve(request.path)

        class FileOrFolder(AdminUrlMixin, object):
            def __init__(s, pk, name):
                s.pk = pk
                s.name = name

            def __str__(s):
                return s.name

            def __repr__(s):
                return "<%s: %s, %r>" % (s.__class__.__name__, s.name, s.urls)

            def _get_admin_namespace(s):
                return match.namespace

            def _get_url_args(s):
                return (s.pk,)

        class File(FileOrFolder):
            icon = "file"

            @cached_property
            def urls(s):
                namespace = s._get_admin_namespace()
                args = s._get_url_args()
                crud = {
                    "detail": default_storage.url(os.path.join(*args)),
                    "delete": reverse("%s:delete" % namespace, args=args),
                }
                return crud

            def _get_url_args(s):
                args = []
                if path:
                    args.append(path)
                if s.pk:
                    args.append(s.pk)
                return args

        class Folder(FileOrFolder):
            icon = "folder"

        object_list = [Folder(d[0], d[1]) for d in directories] + [
            File(f, f) for f in files
        ]
        from collections import namedtuple

        context = {
            "path": path,
            "parent": parentdir,
            "directories": directories,
            "files": files,
            "folder_form": folder_form,
            "file_form": file_form,
            "writable": writable,
            "object_list": object_list,
            "queryset": Page.objects.all(),
            "list_template": first(self.template_path("list.inc.html", "file")),
            "model": namedtuple("file", ()),
        }
        context.update(kwargs)

        templates = self.template_path("list.html", "file")
        return self.render(request, templates, context)

    @staff_login_required_m
    def rm(self, request, path=None, filename=None, *args, **kwargs):
        try:
            directories, files = default_storage.listdir(path or ".")
        except OSError as e:
            logger.exception("path: %r", path)
            raise Http404(e[1])

        class File(object):
            def __init__(self, path):
                self.path = path

            def __str__(self):
                return self.path

            def is_leaf_node(self):
                return True

        class Folder(File):
            def __init__(self, path):
                self.path = os.path.dirname(path)

            def is_leaf_node(self):
                listdir = default_storage.listdir(self.path)
                logger.debug(listdir)
                return not any([item for item in listdir if item])

        fullpath = os.path.join(path or "", filename or "")

        if filename:
            cls, action, tpl = File, default_storage.delete, "file"
            item = os.path.basename(fullpath)
        elif hasattr(default_storage, "rmdir"):
            cls, action, tpl = Folder, default_storage.rmdir, "folder"
            item = os.path.basename(os.path.dirname(fullpath))
        else:
            raise Http404(
                "We can't handle this, as the "
                "backend storage system does not "
                "support removing directories."
            )

        if path is not None:
            parentdir = os.path.normpath(os.path.join(path, "..")) + "/"
        else:
            parentdir = ""

        obj = cls(fullpath)
        if not obj.is_leaf_node():
            messages.error(request, _('The folder "%s" is not empty.') % item)
            return self.redirect(self.reverse("media:edit", kwargs={"path": parentdir}))

        if request.method == "POST":
            action(fullpath)
            messages.success(request, 'Successfully deleted %s "%s"' % (tpl, item))
            if path:
                if cls is Folder:
                    return self.redirect(
                        self.reverse("media:edit", kwargs={"path": parentdir})
                    )
                return self.redirect(self.reverse("media:edit", kwargs={"path": path}))
            return self.redirect(self.reverse("media:list"))

        context = {
            "path": path,
            "parentdir": parentdir,
            "filename": filename,
            "object": obj,
        }
        context.update(kwargs)

        templates = self.template_path("delete_confirm.html", tpl)
        return self.render(request, templates, context)



# -*- coding: utf-8 -*-

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0030_match_referees"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="seasonreferee",
            options={
                "ordering": ("season", "person__last_name", "person__first_name"),
                "verbose_name": "referee",
            },
        ),
        migrations.AlterField(
            model_name="club",
            name="primary",
            field=touchtechnology.common.db.models.ForeignKey(
                blank=True,
                help_text="Appears on the front-end with other club information.",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="competition.Person",
                verbose_name="Primary contact",
            ),
        ),
        migrations.AlterField(
            model_name="club",
            name="slug_locked",
            field=touchtechnology.common.db.models.BooleanField(
                default=False, verbose_name="Slug locked"
            ),
        ),
        migrations.AlterField(
            model_name="clubassociation",
            name="club",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="staff",
                to="competition.Club",
            ),
        ),
        migrations.AlterField(
            model_name="clubassociation",
            name="person",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="competition.Person"
            ),
        ),
        migrations.AlterField(
            model_name="clubrole",
            name="competition",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="club_roles",
                to="competition.Competition",
            ),
        ),
        migrations.AlterField(
            model_name="competition",
            name="enabled",
            field=touchtechnology.common.db.models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name="competition",
            name="slug_locked",
            field=touchtechnology.common.db.models.BooleanField(
                default=False, verbose_name="Slug locked"
            ),
        ),
        migrations.AlterField(
            model_name="division",
            name="draft",
            field=touchtechnology.common.db.models.BooleanField(
                default=False,
                help_text="Marking a division as draft will prevent matches from being visible in the front-end.",
            ),
        ),
        migrations.AlterField(
            model_name="division",
            name="include_forfeits_in_played",
            field=touchtechnology.common.db.models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name="division",
            name="rank_division",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="competition.RankDivision",
            ),
        ),
        migrations.AlterField(
            model_name="division",
            name="season",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="divisions",
                to="competition.Season",
            ),
        ),
        migrations.AlterField(
            model_name="division",
            name="slug_locked",
            field=touchtechnology.common.db.models.BooleanField(
                default=False, verbose_name="Slug locked"
            ),
        ),
        migrations.AlterField(
            model_name="divisionexclusiondate",
            name="date",
            field=models.DateField(),
        ),
        migrations.AlterField(
            model_name="drawformat",
            name="is_final",
            field=touchtechnology.common.db.models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name="ground",
            name="venue",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="grounds",
                to="competition.Venue",
            ),
        ),
        migrations.AlterField(
            model_name="ladderentry",
            name="opponent",
            field=touchtechnology.common.db.models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="competition.Team",
            ),
        ),
        migrations.AlterField(
            model_name="laddersummary",
            name="team",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="ladder_summary",
                to="competition.Team",
            ),
        ),
        migrations.AlterField(
            model_name="match",
            name="away_team",
            field=touchtechnology.common.db.models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="away_games",
                to="competition.Team",
            ),
        ),
        migrations.AlterField(
            model_name="match",
            name="away_team_eval_related",
            field=touchtechnology.common.db.models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="competition.Match",
            ),
        ),
        migrations.AlterField(
            model_name="match",
            name="away_team_undecided",
            field=touchtechnology.common.db.models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="away_games",
                to="competition.UndecidedTeam",
            ),
        ),
        migrations.AlterField(
            model_name="match",
            name="bye_processed",
            field=touchtechnology.common.db.models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name="match",
            name="date",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="match",
            name="datetime",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="match",
            name="forfeit_winner",
            field=touchtechnology.common.db.models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="competition.Team",
            ),
        ),
        migrations.AlterField(
            model_name="match",
            name="home_team",
            field=touchtechnology.common.db.models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="home_games",
                to="competition.Team",
            ),
        ),
        migrations.AlterField(
            model_name="match",
            name="home_team_eval_related",
            field=touchtechnology.common.db.models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="competition.Match",
            ),
        ),
        migrations.AlterField(
            model_name="match",
            name="home_team_undecided",
            field=touchtechnology.common.db.models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="home_games",
                to="competition.UndecidedTeam",
            ),
        ),
        migrations.AlterField(
            model_name="match",
            name="include_in_ladder",
            field=touchtechnology.common.db.models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name="match",
            name="is_bye",
            field=touchtechnology.common.db.models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name="match",
            name="is_forfeit",
            field=touchtechnology.common.db.models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name="match",
            name="is_washout",
            field=touchtechnology.common.db.models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name="match",
            name="stage",
            field=touchtechnology.common.db.models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="matches",
                to="competition.Stage",
            ),
        ),
        migrations.AlterField(
            model_name="match",
            name="stage_group",
            field=touchtechnology.common.db.models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="matches",
                to="competition.StageGroup",
                verbose_name="Pool",
            ),
        ),
        migrations.AlterField(
            model_name="match",
            name="time",
            field=models.TimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="person",
            name="club",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="members",
                to="competition.Club",
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="date_of_birth",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="person",
            name="user",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="place",
            name="slug_locked",
            field=touchtechnology.common.db.models.BooleanField(
                default=False, verbose_name="Slug locked"
            ),
        ),
        migrations.AlterField(
            model_name="rankdivision",
            name="enabled",
            field=touchtechnology.common.db.models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name="rankdivision",
            name="slug_locked",
            field=touchtechnology.common.db.models.BooleanField(
                default=False, verbose_name="Slug locked"
            ),
        ),
        migrations.AlterField(
            model_name="season",
            name="competition",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="seasons",
                to="competition.Competition",
            ),
        ),
        migrations.AlterField(
            model_name="season",
            name="complete",
            field=touchtechnology.common.db.models.BooleanField(
                default=False,
                help_text="Set to indicate this season is in the past. Optimises progression calculations.",
            ),
        ),
        migrations.AlterField(
            model_name="season",
            name="enabled",
            field=touchtechnology.common.db.models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name="season",
            name="forfeit_notifications",
            field=touchtechnology.common.db.models.ManyToManyField(
                blank=True,
                help_text="When a team advises they are forfeiting, notify the opposition team plus these people.",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="season",
            name="mvp_results_public",
            field=models.DateTimeField(
                blank=True,
                help_text="The time when the results of the MVP voting will be made public on the website. Leave blank to show at all times.",
                null=True,
                verbose_name="MVP public at",
            ),
        ),
        migrations.AlterField(
            model_name="season",
            name="slug_locked",
            field=touchtechnology.common.db.models.BooleanField(
                default=False, verbose_name="Slug locked"
            ),
        ),
        migrations.AlterField(
            model_name="season",
            name="start_date",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="season",
            name="statistics",
            field=touchtechnology.common.db.models.BooleanField(
                default=True,
                help_text="Set to No if you do not wish to keep scoring or most valuable player statistics.",
            ),
        ),
        migrations.AlterField(
            model_name="seasonassociation",
            name="club",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="officials",
                to="competition.Club",
            ),
        ),
        migrations.AlterField(
            model_name="seasonassociation",
            name="person",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="competition.Person"
            ),
        ),
        migrations.AlterField(
            model_name="seasonassociation",
            name="season",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="officials",
                to="competition.Season",
            ),
        ),
        migrations.AlterField(
            model_name="seasonexclusiondate",
            name="date",
            field=models.DateField(),
        ),
        migrations.AlterField(
            model_name="seasonmatchtime",
            name="end_date",
            field=models.DateField(blank=True, null=True, verbose_name="Until"),
        ),
        migrations.AlterField(
            model_name="seasonmatchtime",
            name="start",
            field=models.TimeField(),
        ),
        migrations.AlterField(
            model_name="seasonmatchtime",
            name="start_date",
            field=models.DateField(blank=True, null=True, verbose_name="From"),
        ),
        migrations.AlterField(
            model_name="simplescorematchstatistic",
            name="match",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="statistics",
                to="competition.Match",
            ),
        ),
        migrations.AlterField(
            model_name="simplescorematchstatistic",
            name="player",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="statistics",
                to="competition.Person",
            ),
        ),
        migrations.AlterField(
            model_name="stage",
            name="carry_ladder",
            field=touchtechnology.common.db.models.BooleanField(
                default=False,
                help_text="Set this to <b>Yes</b> if this stage should carry over values from the previous stage.",
                verbose_name="Carry over points",
            ),
        ),
        migrations.AlterField(
            model_name="stage",
            name="division",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="stages",
                to="competition.Division",
            ),
        ),
        migrations.AlterField(
            model_name="stage",
            name="follows",
            field=touchtechnology.common.db.models.ForeignKey(
                blank=True,
                help_text="When progressing teams into this stage, which earlier stage should be used for determining positions.<br>Default is the immediately preceeding stage.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="preceeds",
                to="competition.Stage",
            ),
        ),
        migrations.AlterField(
            model_name="stage",
            name="keep_ladder",
            field=touchtechnology.common.db.models.BooleanField(
                default=True,
                help_text="Set this to <b>No</b> if this stage does not need to keep a competition ladder.<br>Usually set to No for a Final Series or a Knockout stage.",
                verbose_name="Keep a ladder",
            ),
        ),
        migrations.AlterField(
            model_name="stage",
            name="keep_mvp",
            field=touchtechnology.common.db.models.BooleanField(
                default=True,
                help_text="Set this to <b>No</b> if this stage does not need to keep track of MVP points.<br>Usually set to No for a Final Series.",
                verbose_name="Keep MVP stats",
            ),
        ),
        migrations.AlterField(
            model_name="stage",
            name="scale_group_points",
            field=touchtechnology.common.db.models.BooleanField(
                default=False,
                help_text="In stages with multiple pools, adjust points in the smaller groups to compensate for the reduced opportunity to score points.<br>You <strong>should</strong> also set 0 points for Bye matches.",
            ),
        ),
        migrations.AlterField(
            model_name="stage",
            name="slug_locked",
            field=touchtechnology.common.db.models.BooleanField(
                default=False, verbose_name="Slug locked"
            ),
        ),
        migrations.AlterField(
            model_name="stagegroup",
            name="carry_ladder",
            field=touchtechnology.common.db.models.BooleanField(
                default=False,
                help_text="Set this to <b>Yes</b> if the ladder for this pool should carry over values from the previous stage.<br>Will only apply for matches played against teams that are now in this group.",
                verbose_name="Carry over points",
            ),
        ),
        migrations.AlterField(
            model_name="stagegroup",
            name="slug_locked",
            field=touchtechnology.common.db.models.BooleanField(
                default=False, verbose_name="Slug locked"
            ),
        ),
        migrations.AlterField(
            model_name="stagegroup",
            name="stage",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="pools",
                to="competition.Stage",
            ),
        ),
        migrations.AlterField(
            model_name="team",
            name="division",
            field=touchtechnology.common.db.models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="teams",
                to="competition.Division",
            ),
        ),
        migrations.AlterField(
            model_name="team",
            name="names_locked",
            field=touchtechnology.common.db.models.BooleanField(
                default=False,
                help_text="When the team name is locked, the team manager will not be able to change their team name.<br>As a tournament manager you can always change the names.",
            ),
        ),
        migrations.AlterField(
            model_name="team",
            name="rank_division",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="competition.RankDivision",
            ),
        ),
        migrations.AlterField(
            model_name="team",
            name="slug_locked",
            field=touchtechnology.common.db.models.BooleanField(
                default=False, verbose_name="Slug locked"
            ),
        ),
        migrations.AlterField(
            model_name="team",
            name="timeslots_after",
            field=models.TimeField(
                blank=True,
                help_text="Specify the earliest time that this team can play. Leave blank for no preference.",
                null=True,
                verbose_name="Start after",
            ),
        ),
        migrations.AlterField(
            model_name="team",
            name="timeslots_before",
            field=models.TimeField(
                blank=True,
                help_text="Specify the latest time that this team can play. Leave blank for no preference.",
                null=True,
                verbose_name="Start before",
            ),
        ),
        migrations.AlterField(
            model_name="teamassociation",
            name="is_player",
            field=touchtechnology.common.db.models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name="teamassociation",
            name="person",
            field=touchtechnology.common.db.models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="competition.Person",
            ),
        ),
        migrations.AlterField(
            model_name="teamassociation",
            name="team",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="people",
                to="competition.Team",
            ),
        ),
        migrations.AlterField(
            model_name="teamrole",
            name="competition",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="team_roles",
                to="competition.Competition",
            ),
        ),
        migrations.AlterField(
            model_name="undecidedteam",
            name="stage",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="undecided_teams",
                to="competition.Stage",
            ),
        ),
        migrations.AlterField(
            model_name="venue",
            name="season",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="venues",
                to="competition.Season",
            ),
        ),
    ]



import collections
import logging

from django.db.models import Count, Q
from django.utils import timezone
from django.utils.functional import lazy
from django.utils.translation import gettext_lazy as _

from touchtechnology.admin.base import DashboardWidget
from tournamentcontrol.competition.models import Match, Stage, Team
from tournamentcontrol.competition.utils import (
    legitimate_bye_match,
    team_needs_progressing,
)

logger = logging.getLogger(__name__)


def matches_require_basic_results(now=None, matches=None):
    if now is None:
        now = timezone.now()
        now = now.replace(second=0, microsecond=0)

    if timezone.is_naive(now):
        timezone.make_aware(now, timezone.get_default_timezone())

    # If not provided up front, build a base queryset of matches that take
    # place prior to "now".
    if matches is None:
        try:
            last_bye_date = (
                Match.objects.filter(datetime__lte=now).dates("date", "day").latest()
            )
        except Match.DoesNotExist:
            last_bye_date = now.date()
        matches = Match.objects.filter(
            Q(datetime__lte=now)
            | Q(is_bye=True, bye_processed=False, date__lte=last_bye_date),
            stage__division__season__complete=False,
        )

    return matches.filter(
        home_team_score=None,
        away_team_score=None,
        is_washout=False,
    ).select_related(
        "stage__division__season__competition",
        "play_at",
        "home_team__club",
        "home_team__division",
        "away_team__club",
        "away_team__division",
    )


def matches_require_details_results(matches=None, include_forfeits=False):
    # If not provided up front, build a base queryset of all matches
    if matches is None:
        matches = Match.objects.filter(
            is_forfeit=False,
            stage__division__season__complete=False,
        )

    matches = matches.filter(
        stage__division__season__statistics=True,
        statistics__isnull=True,
        home_team__isnull=False,
        home_team_score__isnull=False,
        away_team__isnull=False,
        away_team_score__isnull=False,
    )

    if not include_forfeits:
        matches = matches.filter(is_forfeit=False)

    return matches.select_related(
        "stage__division__season__competition",
        "play_at",
        "home_team__club",
        "home_team__division",
        "away_team__club",
        "away_team__division",
    )


def matches_require_progression():
    matches = Match.objects.filter(team_needs_progressing).exclude(legitimate_bye_match)
    matches = matches.order_by("stage__division", "stage")

    return matches.select_related(
        "stage__division__season__competition",
        "play_at",
        "home_team__club",
        "home_team__division",
        "away_team__club",
        "away_team__division",
    )


def matches_progression_possible():
    matches = matches_require_progression()

    # cache which stages have all their matches played or byes marked played
    stages = Stage.objects.filter(matches__in=matches).distinct()
    score_entered = Q(home_team_score__isnull=False, away_team_score__isnull=False)
    played_bye = Q(is_bye=True, bye_processed=True)
    is_washout = Q(is_washout=True)
    _stage_cache = {}
    for stage in stages:
        f = stage.comes_after
        try:
            _stage_cache[stage.pk] = f.matches.exclude(
                score_entered | played_bye | is_washout
            ).count()
        except AttributeError:
            _stage_cache[stage.pk] = 0

    def _can_evaluate(match):
        """
        If we can actually evaluate a Team instance for assignment into
        the `home_team` OR `away_team` field, then this is a match we
        want to know about. Not possible to do directly in a QuerySet as
        there is too much logic on the models, so we need to attempt this
        for every match that could possibly be progressed.
        """
        # If there are any unplayed or unprocessed matches in the preceding
        # stage, then we do not want to consider this match as viable for
        # progression.
        logger.debug(
            "Evaluate %r - home_team=%s home_team_eval=%s home_team_eval_related=%s",
            match,
            match.home_team,
            match.home_team_eval,
            match.home_team_eval_related,
        )
        logger.debug(
            "Evaluate %r - away_team=%s away_team_eval=%s away_team_eval_related=%s",
            match,
            match.away_team,
            match.away_team_eval,
            match.away_team_eval_related,
        )
        if _stage_cache[match.stage_id]:
            logger.warning(
                "Can't eval %s from stage %r (%s matches)",
                match,
                match.stage,
                _stage_cache[match.stage_id],
            )
            return False

        # If the home or away team can be determined we would want to progress
        # the match. This should only catch Px, GxPx, Wx, Lx - for
        # UndecidedTeam cases we need to catch them all together.
        home_team, away_team = match.eval()
        if match.home_team is None and isinstance(home_team, Team):
            return True
        elif match.away_team is None and isinstance(away_team, Team):
            return True

        if match.stage.undecided_teams.exists() and _stage_cache[match.stage_id] == 0:
            logger.warning(
                "Found undecided teams and no matches left for %s in stage %r",
                match,
                match.stage,
            )
            return True

        logger.warning(
            "Can't evaluate %r - home_team=%r away_team=%r", match, home_team, away_team
        )
        return False

    matches = [m for m in matches if _can_evaluate(m)]

    return matches


def stages_require_progression():
    matches = matches_progression_possible()

    # Turn the list of matches with progressions into a nested OrderedDict
    # for use in the template. Use of OrderedDict means our ordering set
    # above will not be lost.
    stages = collections.OrderedDict()
    for m in matches:
        stages.setdefault(m.stage.division, collections.OrderedDict()).setdefault(
            m.stage, []
        ).append(m)

    return stages


lazy_stages_require_progression = lazy(stages_require_progression, dict)


#
# Dashboard Widgets
#


class BasicResultWidget(DashboardWidget):
    verbose_name = _("Awaiting Scores")
    template = "tournamentcontrol/competition/admin/widgets/results/basic.html"

    def _get_context(self):
        matches = matches_require_basic_results()
        matches = matches.order_by("date", "time")

        dates = matches.values_list(
            "stage__division__season__competition", "stage__division__season", "date"
        ).distinct()
        times = matches.values_list(
            "stage__division__season__competition",
            "stage__division__season",
            "date",
            "time",
        ).distinct()

        # Construct an interim data structure
        data = collections.OrderedDict()
        for competition, season, date, time in times:
            key = (competition, season, date)
            data.setdefault(key, []).append(time)

        # Remove any None values from the list of times if there are any
        # real times in the list. If not, we'll keep it so our template
        # can iterate the "loop" at least once.
        dates_times = []
        for key, time_list in data.items():
            if [t for t in time_list if t] and None in time_list:
                time_list.remove(None)
            for time in time_list:
                dates_times.append(key + (time,))

        context = {"matches": matches, "dates": dates, "dates_times": dates_times}
        return context


class ProgressStageWidget(DashboardWidget):
    verbose_name = _("Progress Teams")
    template = "tournamentcontrol/competition/admin/widgets/progress/stages.html"

    def _get_context(self):
        stages = lazy_stages_require_progression()

        context = {"stages": stages}
        return context


class DetailResultWidget(DashboardWidget):
    verbose_name = _("Awaiting Detailed Results")
    template = "tournamentcontrol/competition/admin/widgets/results/detailed.html"

    @classmethod
    def show(cls):
        matches = matches_require_details_results()
        return bool(matches.count())

    @property
    def matches(self):
        if not hasattr(self, "_matches"):
            self._matches = matches_require_details_results()
        return self._matches

    def _get_context(self):
        context = {"matches": self.matches}
        return context


class MostValuableWidget(DashboardWidget):
    verbose_name = _("Awaiting MVP Points")
    template = "tournamentcontrol/competition/admin/widgets/results/detailed.html"

    @property
    def matches(self):
        sql = """
            SELECT DISTINCT
                m.id, m.home_team_id, m.away_team_id, m.stage_id, m.date,
                m.datetime, m.label, m.round
            FROM
                competition_simplescorematchstatistic s
            JOIN
                competition_teamassociation t ON (s.player_id = t.person_id)
            JOIN
                  competition_match m
                ON (
                    m.id = s.match_id
                  AND (
                      m.home_team_id = t.team_id
                    OR
                      m.away_team_id = t.team_id
                  )
                )
            JOIN
                competition_stage g ON (m.stage_id = g.id)
            JOIN
                competition_division d ON (g.division_id = d.id)
            JOIN
                competition_season se ON (d.season_id = se.id)
            WHERE
                  t.is_player
                AND
                  g.keep_mvp
                AND
                  (NOT se.complete OR se.mvp_results_public IS NULL)
            GROUP BY
                m.id, team_id, m.home_team_id, m.away_team_id, m.stage_id,
                m.date, m.datetime, m.label, m.round
            HAVING
                SUM(mvp) < %s
        """
        # FIXME should not be a static value
        return Match.objects.raw(sql, params=(1,))

    def _get_context(self):
        context = {"matches": self.matches}
        return context


class ScoresheetWidget(DashboardWidget):
    verbose_name = _("Season Reports")
    template = "tournamentcontrol/competition/admin/widgets/scoresheets.html"

    def _get_context(self):
        stages = Stage.objects.annotate(p=Count("matches_needing_printing")).filter(
            p__gt=0
        )
        context = {"stages": stages}
        return context


class ReportWidget(DashboardWidget):
    verbose_name = _("Reports")
    template = "tournamentcontrol/competition/admin/widgets/reports.html"

    def _get_context(self):
        context = {}
        return context



# -*- coding: utf-8 -*-

from django.db import migrations, models


def set_zero_to_none(apps, schema_editor):
    SimpleScoreMatchStatistic = apps.get_model(
        "competition", "SimpleScoreMatchStatistic"
    )
    SimpleScoreMatchStatistic.objects.filter(mvp=0).update(mvp=None)
    SimpleScoreMatchStatistic.objects.filter(points=0).update(points=None)


def set_none_to_zero(apps, schema_editor):
    SimpleScoreMatchStatistic = apps.get_model(
        "competition", "SimpleScoreMatchStatistic"
    )
    SimpleScoreMatchStatistic.objects.filter(mvp__isnull=True).update(mvp=0)
    SimpleScoreMatchStatistic.objects.filter(points__isnull=True).update(points=0)


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0007_match_videos"),
    ]

    operations = [
        migrations.AlterField(
            model_name="simplescorematchstatistic",
            name="mvp",
            field=models.SmallIntegerField(null=True, blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="simplescorematchstatistic",
            name="points",
            field=models.SmallIntegerField(null=True, blank=True),
            preserve_default=True,
        ),
        migrations.RunPython(set_zero_to_none, set_none_to_zero),
    ]



"""
LiveScore / LiveStream integration API resources.

This is not considered official API and should not be relied upon - it could change
at any time with changing needs.
"""

from rest_framework import serializers, viewsets

from tournamentcontrol.competition import models


class PlayerSerializer(serializers.Serializer):
    id = serializers.IntegerField(read_only=True)
    division = serializers.ReadOnlyField(source="team.division.title")
    team = serializers.ReadOnlyField(source="team.title")
    name = serializers.SerializerMethodField(read_only=True)
    number = serializers.IntegerField(read_only=True)

    def get_name(self, obj):
        return f"{obj.person.first_name} {obj.person.last_name}".upper()


class PersonViewSet(viewsets.ModelViewSet):
    """
    This endpoint is used by the 247.tv developed application to seed it's
    list of players for use with the YouTube Event API for live streams.
    """

    lookup_field = "id"
    serializer_class = PlayerSerializer

    def get_queryset(self):
        return (
            models.TeamAssociation.objects.filter(
                team__division__season__slug=self.kwargs["season_slug"],
                team__division__season__competition__slug=self.kwargs[
                    "competition_slug"
                ],
                is_player=True,
            )
            .select_related("team__division", "team", "person")
            .order_by("team__division", "team", "number")
        )



# -*- coding: utf-8 -*-

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0024_auto_20170203_0619"),
    ]

    operations = [
        migrations.AlterField(
            model_name="division",
            name="sportingpulse_url",
            field=models.URLField(
                blank=True, editable=False, max_length=1024, null=True
            ),
        ),
        migrations.AlterUniqueTogether(
            name="division",
            unique_together=set([("slug", "season"), ("title", "season")]),
        ),
        migrations.AlterUniqueTogether(
            name="season",
            unique_together=set([("slug", "competition"), ("title", "competition")]),
        ),
        migrations.AlterUniqueTogether(
            name="stage",
            unique_together=set([("slug", "division"), ("title", "division")]),
        ),
    ]



# -*- coding: utf-8 -*-

import django.db.models.deletion
from django.db import migrations, models

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0017_ranking_feature"),
    ]
    operations = [
        migrations.AlterModelOptions(
            name="clubassociation",
            options={
                "ordering": ("person__last_name", "person__first_name"),
                "verbose_name": "Official",
                "verbose_name_plural": "Officials",
            },
        ),
        migrations.AlterModelOptions(
            name="competition",
            options={},
        ),
        migrations.AlterModelOptions(
            name="division",
            options={},
        ),
        migrations.AlterModelOptions(
            name="divisionexclusiondate",
            options={"verbose_name": "exclusion date"},
        ),
        migrations.AlterModelOptions(
            name="drawformat",
            options={"ordering": ("teams", "name")},
        ),
        migrations.AlterModelOptions(
            name="match",
            options={
                "get_latest_by": "datetime",
                "ordering": (
                    "date",
                    "stage",
                    "round",
                    "is_bye",
                    "time",
                    "play_at__ground__order",
                    "pk",
                ),
                "verbose_name_plural": "matches",
            },
        ),
        migrations.AlterModelOptions(
            name="person",
            options={
                "ordering": ("last_name", "first_name"),
                "verbose_name_plural": "people",
            },
        ),
        migrations.AlterModelOptions(
            name="season",
            options={},
        ),
        migrations.AlterModelOptions(
            name="seasonexclusiondate",
            options={
                "ordering": ("date",),
                "verbose_name": "exclusion date",
            },
        ),
        migrations.AlterModelOptions(
            name="seasonmatchtime",
            options={
                "verbose_name": "time slot",
            },
        ),
        migrations.AlterModelOptions(
            name="simplescorematchstatistic",
            options={
                "get_latest_by": "match",
                "ordering": ("match", "number"),
            },
        ),
        migrations.AlterModelOptions(
            name="stagegroup",
            options={
                "ordering": ("order",),
                "verbose_name": "pool",
            },
        ),
        migrations.AlterModelOptions(
            name="team",
            options={
                "ordering": (
                    "-division__season__start_date",
                    "division__order",
                    "stage_group__order",
                    "order",
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="teamassociation",
            options={
                "ordering": (
                    "-is_player",
                    "number",
                    "person__last_name",
                    "person__first_name",
                ),
                "verbose_name": "linked person",
                "verbose_name_plural": "linked people",
            },
        ),
        migrations.AlterModelOptions(
            name="undecidedteam",
            options={"ordering": ("stage_group", "formula")},
        ),
        migrations.AlterField(
            model_name="club",
            name="primary",
            field=touchtechnology.common.db.models.ForeignKey(
                to="competition.Person",
                verbose_name="Primary contact",
                related_name="+",
                null=True,
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                help_text="Appears on the front-end with other club " "information.",
            ),
        ),
        migrations.AlterField(
            model_name="division",
            name="draft",
            field=models.BooleanField(
                default=False,
                help_text="Marking a division as draft will prevent matches "
                "from being visible in the front-end.",
            ),
        ),
        migrations.AlterField(
            model_name="division",
            name="games_per_day",
            field=models.SmallIntegerField(
                null=True,
                help_text="In Tournament mode, specify how many matches per "
                "day should be scheduled by the automatic draw "
                "generator.",
            ),
        ),
        migrations.AlterField(
            model_name="division",
            name="include_forfeits_in_played",
            field=models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name="match",
            name="away_team_eval_related",
            field=touchtechnology.common.db.models.ForeignKey(
                to="competition.Match",
                related_name="+",
                null=True,
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
        ),
        migrations.AlterField(
            model_name="match",
            name="forfeit_winner",
            field=touchtechnology.common.db.models.ForeignKey(
                to="competition.Team",
                related_name="+",
                null=True,
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
        ),
        migrations.AlterField(
            model_name="match",
            name="home_team_eval_related",
            field=touchtechnology.common.db.models.ForeignKey(
                to="competition.Match",
                related_name="+",
                null=True,
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="email",
            field=models.EmailField(max_length=254, blank=True),
        ),
        migrations.AlterField(
            model_name="person",
            name="first_name",
            field=models.CharField(max_length=50),
        ),
        migrations.AlterField(
            model_name="person",
            name="last_name",
            field=models.CharField(max_length=50),
        ),
        migrations.AlterField(
            model_name="season",
            name="mvp_results_public",
            field=touchtechnology.common.db.models.DateTimeField(
                verbose_name="MVP public at",
                null=True,
                blank=True,
                help_text="The time when the results of the MVP voting will "
                "be made public on the website. Leave blank to show "
                "at all times.",
            ),
        ),
        migrations.AlterField(
            model_name="simplescorematchstatistic",
            name="mvp",
            field=models.SmallIntegerField(
                verbose_name="MVP",
                null=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="simplescorematchstatistic",
            name="points",
            field=models.SmallIntegerField(
                verbose_name="Points",
                null=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="stage",
            name="carry_ladder",
            field=models.BooleanField(
                verbose_name="Carry over points",
                default=False,
                help_text="Set this to <b>Yes</b> if this stage should carry "
                "over values from the previous stage.",
            ),
        ),
        migrations.AlterField(
            model_name="stage",
            name="follows",
            field=touchtechnology.common.db.models.ForeignKey(
                to="competition.Stage",
                related_name="preceeds",
                on_delete=django.db.models.deletion.SET_NULL,
                null=True,
                blank=True,
                help_text="When progressing teams into this stage, which "
                "earlier stage should be used for determining "
                "positions.<br />Default is the immediately "
                "preceeding stage.",
            ),
        ),
        migrations.AlterField(
            model_name="stage",
            name="keep_ladder",
            field=models.BooleanField(
                verbose_name="Keep a ladder",
                default=True,
                help_text="Set this to <b>No</b> if this stage does not need "
                "to keep a competition ladder.<br />Usually set to "
                "No for a Final Series or a Knockout stage.",
            ),
        ),
        migrations.AlterField(
            model_name="stage",
            name="keep_mvp",
            field=models.BooleanField(
                verbose_name="Keep MVP stats",
                default=True,
                help_text="Set this to <b>No</b> if this stage does not need "
                "to keep track of MVP points.<br />Usually set to "
                "No for a Final Series.",
            ),
        ),
        migrations.AlterField(
            model_name="stage",
            name="scale_group_points",
            field=models.BooleanField(
                default=False,
                help_text="In stages with multiple pools, adjust points in "
                "the smaller groups to compensate for the reduced "
                "opportunity to score points.<br />"
                "You <strong>should</strong> also set 0 points for "
                "Bye matches.",
            ),
        ),
        migrations.AlterField(
            model_name="stagegroup",
            name="carry_ladder",
            field=models.BooleanField(
                verbose_name="Carry over points",
                default=False,
                help_text="Set this to <b>Yes</b> if the ladder for this pool "
                "should carry over values from the previous "
                "stage.<br />"
                "Will only apply for matches played against teams "
                "that are now in this group.",
            ),
        ),
        migrations.AlterField(
            model_name="team",
            name="names_locked",
            field=models.BooleanField(
                default=False,
                help_text="When the team name is locked, the team manager "
                "will not be able to change their team name.<br />"
                "As a tournament manager you can always change the "
                "names.",
            ),
        ),
    ]



import os
import socket

from django.utils import timezone

from touchtechnology.common.forms.tz import SelectTimezoneForm


def env(request):
    return {"SITE_ENV": os.environ.get("SITE_ENV", "dev")}


def query_string(request):
    return {"QUERY_STRING": request.META.get("QUERY_STRING")}


def site(request):
    return {"hostname": socket.gethostname()}


def tz(request):
    return {
        "select_timezone_form": SelectTimezoneForm(),
        "TIME_ZONE": timezone.get_current_timezone_name(),
    }



import os.path


class TemplateChoiceIterator(object):
    def __init__(self, field):
        self.field = field

    def __iter__(self):
        folder = os.path.join(self.field.template_base, self.field.template_folder)
        choices = []
        if self.field.recursive:
            for root, dirs, files in os.walk(folder):
                for f in files:
                    if self.field.match is None or self.field.match_re.search(f):
                        f = os.path.join(root, f)
                        base = f.replace(self.field.template_base, "")
                        choices.append((base, f.replace(folder, "", 1)))
        else:
            try:
                for f in os.listdir(folder):
                    full_file = os.path.join(folder, f)
                    if os.path.isfile(full_file) and (
                        self.field.match is None or self.field.match_re.search(f)
                    ):
                        choices.append((full_file, f))
            except OSError:
                pass

        if not self.field.required:
            yield ("", self.field.empty_label)

        if choices:
            yield ("Static template", choices)

    def choice(self, obj):
        path = obj.path.replace(self.field.template_folder, "", 1)
        return (obj.path, "%s (%s)" % (path, obj.name))



from django.conf import settings
from django.core.checks import Error, register


@register()
def check_use_tz_enabled(app_configs, **kwargs):
    """
    Ensure that USE_TZ is turned on.
    """
    errors = []

    if not settings.USE_TZ:
        errors.append(
            Error(
                "USE_TZ must be set to True",
                hint="Set USE_TZ = True in your settings file.",
                id="touchtechnology.common.E001",
            )
        )

    return errors



# -*- coding: utf-8 -*-

import django.db.models.deletion
from django.db import migrations, models

import touchtechnology.common.db.models
import tournamentcontrol.competition.models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0028_drop_matchvideo"),
    ]

    operations = [
        migrations.CreateModel(
            name="SeasonReferee",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "ordering": (
                    "club",
                    "season",
                    "person__last_name",
                    "person__first_name",
                ),
            },
            bases=(tournamentcontrol.competition.models.AdminUrlMixin, models.Model),
        ),
        migrations.AddField(
            model_name="seasonreferee",
            name="club",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="referees",
                to="competition.Club",
            ),
        ),
        migrations.AddField(
            model_name="seasonreferee",
            name="person",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="competition.Person"
            ),
        ),
        migrations.AddField(
            model_name="seasonreferee",
            name="season",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="referees",
                to="competition.Season",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="seasonreferee",
            unique_together=set([("season", "person")]),
        ),
    ]



from django.apps import AppConfig


class CommonConfig(AppConfig):
    name = "touchtechnology.common"
    verbose_name = "Touch Technology Common"

    def ready(self):
        from . import checks  # noqa



from django.urls import include, path

urlpatterns = [
    path(
        "v1/",
        include(
            ("tournamentcontrol.competition.rest.v1.urls", "api"),
            namespace="v1",
        ),
    ),
    path("api-auth/", include("rest_framework.urls")),
]



# Generated by Django 4.2.20 on 2025-05-03 06:12

import timezone_field.fields
from django.db import migrations, models

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0049_match_live_stream_thumbnail"),
    ]

    operations = [
        migrations.AlterField(
            model_name="place",
            name="timezone",
            field=timezone_field.fields.TimeZoneField(
                blank=True, max_length=50, null=True, use_pytz=False
            ),
        ),
        migrations.AlterField(
            model_name="season",
            name="live_stream_thumbnail",
            field=models.URLField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="season",
            name="timezone",
            field=timezone_field.fields.TimeZoneField(
                blank=True, max_length=50, null=True, use_pytz=False
            ),
        ),
        migrations.AlterField(
            model_name="team",
            name="team_clashes",
            field=touchtechnology.common.db.models.ManyToManyField(
                blank=True,
                help_text="Select any teams that must not play at the same time.",
                to="competition.team",
                verbose_name="Don't clash",
            ),
        ),
    ]



# -*- coding: utf-8 -*-

import django.core.validators
from django.db import migrations, models


def prefix_with_hash(apps, schema_editor):
    Season = apps.get_model("competition", "Season")
    for season in Season.objects.exclude(hashtag__isnull=True):
        season.hashtag = f"#{season.hashtag}"
        season.save()


def remove_hash_prefix(apps, schema_editor):
    Season = apps.get_model("competition", "Season")
    for season in Season.objects.filter(hashtag__startswith="#"):
        season.hashtag = season.hashtag[1:]
        season.save()


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0011_custom_twitter_ladderpoints_fields"),
    ]

    operations = [
        migrations.AlterField(
            model_name="season",
            name="hashtag",
            field=models.CharField(
                validators=[
                    django.core.validators.RegexValidator(
                        b"^(?:#)(\\w+)$",
                        "Enter a valid value. Make sure you include the # symbol.",
                    )
                ],
                max_length=30,
                blank=True,
                help_text="Your official <em>hash tag</em> for social media promotions.",
                null=True,
                verbose_name="Hash Tag",
            ),
        ),
        migrations.RunPython(prefix_with_hash, remove_hash_prefix),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("news", "0004_model_definition_sync"),
    ]

    operations = [
        migrations.AlterField(
            model_name="articlecontent",
            name="label",
            field=models.SlugField(
                choices=[("copy", "Copy")], max_length=20, verbose_name="CSS class"
            ),
        ),
    ]



from django.contrib.auth.models import Group
from test_plus import TestCase as PlusTestCase

from touchtechnology.common.tests import factories


class TestCase(PlusTestCase):

    def setUp(self):
        super(TestCase, self).setUp()
        self.staff = factories.UserFactory.create(is_staff=True)
        self.superuser = factories.UserFactory.create(is_staff=True, is_superuser=True)

    def get_generic_200(self, url_name, user, in_context, template, **kw):
        self.assertLoginRequired(url_name, **kw)
        with self.login(user):
            res = self.get(url_name, **kw)
            self.response_200()
            for key in in_context:
                self.assertInContext(key)
            self.assertTemplateUsed(res, template)


class AuthSiteUserTests(TestCase):

    def test_index(self):
        url_name = "admin:auth:index"

        self.assertLoginRequired(url_name)

        with self.login(self.superuser):
            self.get(url_name)
            self.response_302()

    def test_user_list(self):
        self.get_generic_200(
            "admin:auth:users:list",
            self.superuser,
            ["model", "object_list"],
            "touchtechnology/admin/list.html",
        )

    def test_user_create(self):
        self.get_generic_200(
            "admin:auth:users:add",
            self.superuser,
            ["model", "form"],
            "touchtechnology/admin/edit.html",
        )

    def test_user_edit(self):
        self.get_generic_200(
            "admin:auth:users:edit",
            self.superuser,
            ["model", "form", "object"],
            "touchtechnology/admin/edit.html",
            pk=1,
        )

    def test_user_perms(self):
        self.get_generic_200(
            "admin:auth:users:perms",
            self.superuser,
            ["model", "formset"],
            "touchtechnology/admin/permissions.html",
            pk=1,
        )

    def test_user_delete(self):
        url_name = "admin:auth:users:delete"

        with self.login(self.superuser):
            self.get(url_name, pk=1)
            self.response_405()

            self.post(url_name, pk=1)
            self.response_302()


class AuthSiteGroupTests(TestCase):

    def setUp(self):
        super(AuthSiteGroupTests, self).setUp()
        self.group_instance = Group.objects.create(name="Group")

    def test_group_list(self):
        self.get_generic_200(
            "admin:auth:groups:list",
            self.superuser,
            ["model", "object_list"],
            "touchtechnology/admin/list.html",
        )

    def test_group_create(self):
        self.get_generic_200(
            "admin:auth:groups:add",
            self.superuser,
            ["model", "form"],
            "touchtechnology/admin/edit.html",
        )

    def test_group_edit(self):
        self.get_generic_200(
            "admin:auth:groups:edit",
            self.superuser,
            ["model", "form", "object"],
            "touchtechnology/admin/edit.html",
            pk=self.group_instance.pk,
        )

    def test_group_perms(self):
        self.get_generic_200(
            "admin:auth:groups:perms",
            self.superuser,
            ["model", "formset"],
            "touchtechnology/admin/permissions.html",
            pk=self.group_instance.pk,
        )

    def test_group_delete(self):
        url_name = "admin:auth:groups:delete"

        with self.login(self.superuser):
            self.get(url_name, pk=self.group_instance.pk)
            self.response_405()

            self.post(url_name, pk=self.group_instance.pk)
            self.response_302()



import logging

logger = logging.getLogger(__name__)

NAME = "Competitions"
INSTALL = (
    "CompetitionSite",
    "MultiCompetitionSite",
    "RegistrationSite",
    "TournamentCalculatorSite",
    "RankingSite",
)

__version__ = "2.4.14"

logger.debug('"%s"/"%s"' % (NAME, __version__))

default_app_config = "tournamentcontrol.competition.apps.CompetitionConfig"



from django.conf import settings
from django.db.models import Case, ExpressionWrapper, F, FloatField, Func, When
from django.db.models.query import QuerySet
from django.utils import timezone
from django.utils.module_loading import import_string


class DivisionQuerySet(QuerySet):
    """
    QuerySet to allow divisions to be easily filtered by their draft status.
    """

    def public(self):
        return self.filter(draft=False)


class StageQuerySet(QuerySet):
    def with_ladder(self):
        return self.filter(keep_ladder=True, ladder_summary__isnull=False).distinct()


class MatchQuerySet(QuerySet):
    def future(self, date=None):
        if date is None:
            date = timezone.now().date()
        return self.filter(date__gte=date)

    def playable(self):
        return self.exclude(is_bye=True)

    def videos(self):
        return self.filter(videos__isnull=False).order_by("datetime").distinct()

    def _rank_importance(self):
        """
        Find the rank_importance of a match based on the competition hierarchy.
        """
        return self.annotate(
            importance=Case(
                When(rank_importance__isnull=False, then=F("rank_importance")),
                When(
                    stage_group__rank_importance__isnull=False,
                    then=F("stage_group__rank_importance"),
                ),
                When(
                    stage__rank_importance__isnull=False,
                    then=F("stage__rank_importance"),
                ),
                When(
                    stage__division__rank_importance__isnull=False,
                    then=F("stage__division__rank_importance"),
                ),
                When(
                    stage__division__season__rank_importance__isnull=False,
                    then=F("stage__division__season__rank_importance"),
                ),
                When(
                    stage__division__season__competition__rank_importance__isnull=False,
                    then=F("stage__division__season__competition__rank_importance"),
                ),
            ),
        )


class LadderEntryQuerySet(QuerySet):
    def _all(self):
        qs = self.annotate(
            diff=F("score_for") - F("score_against"),
            margin=Func(
                F("score_for") - F("score_against"),
                function="ABS",
                output_field=FloatField(),
            ),
        )

        qs = qs.select_related("team__club")

        qs = qs.annotate(
            division=Case(
                When(team__rank_division__isnull=False, then=F("team__rank_division")),
                When(
                    team__division__rank_division__isnull=False,
                    then=F("team__division__rank_division"),
                ),
            ),
            opponent_division=Case(
                When(
                    opponent__rank_division__isnull=False,
                    then=F("opponent__rank_division"),
                ),
                When(
                    opponent__division__rank_division__isnull=False,
                    then=F("opponent__division__rank_division"),
                ),
            ),
        )

        qs = qs.annotate(
            importance=Case(
                When(
                    match__rank_importance__isnull=False,
                    then=F("match__rank_importance"),
                ),
                When(
                    match__stage_group__rank_importance__isnull=False,
                    then=F("match__stage_group__rank_importance"),
                ),
                When(
                    match__stage__rank_importance__isnull=False,
                    then=F("match__stage__rank_importance"),
                ),
                When(
                    match__stage__division__rank_importance__isnull=False,
                    then=F("match__stage__division__rank_importance"),
                ),
                When(
                    match__stage__division__season__rank_importance__isnull=False,
                    then=F("match__stage__division__season__rank_importance"),
                ),
                When(
                    match__stage__division__season__competition__rank_importance__isnull=False,
                    then=F(
                        "match__stage__division__season__competition__rank_importance"
                    ),
                ),
                output_field=FloatField(),
            ),
        )

        RANK_POINTS_FUNC = getattr(
            settings,
            "TOURNAMENTCONTROL_RANK_POINTS_FUNC",
            "tournamentcontrol.competition.rank.points_func",
        )
        rank_points = import_string(RANK_POINTS_FUNC)
        qs = qs.annotate(
            rank_points=ExpressionWrapper(
                rank_points() * F("importance"),
                output_field=FloatField(),
            ),
        )

        return qs


class StatisticQuerySet(QuerySet):
    def played(self):
        return self.exclude(played=0)



# -*- coding: utf-8 -*-

import uuid

from django.db import migrations


def initial_uuid(apps, schema_editor):
    Match = apps.get_model("competition", "Match")

    for m in Match.objects.select_related("stage__division"):
        m.uuid = uuid.UUID(
            "{00000000-%04x-%04x-%04x-000000000000}"
            % (m.stage.division.pk, m.stage.pk, m.pk)
        )
        m.save()


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0023_competition__add__match_uuid"),
    ]

    operations = [
        migrations.RunPython(initial_uuid, reverse_code=migrations.RunPython.noop),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations


def set_draft_false(apps, schema_editor):
    Division = apps.get_model("competition", "Division")
    Division.objects.update(draft=False)


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0002_division_draft"),
    ]

    operations = [
        migrations.RunPython(set_draft_false, set_draft_false),
    ]
